# Functions to estimate the applicability domain: 

# ADomain (based on Sahigara, F.; Ballabio, D.; Todeschini, R.; # Consonni, V. Defining a novel k-nearest neighbours
# approach to assess the applicability domain of a QSAR model for reliable predictions. J. Cheminformatics 2013, 5, 27)

ADomain <- function(traindataset, testdataset, model_colnames, method) { # method may be "Chi-squared", "euclidean", "gower"
  
  n_train <- nrow(traindataset)
  kvalue <- round(n_train^(1/3))
  kvalue
  
  # install.packages("proxy")
  library(proxy)
  # pr_DB$get_entries()
  distset <- dist(traindataset[, model_colnames], method=method, upper=TRUE) # "Chi-squared", "euclidean", "gower"
  # model_colnames should not include the dependent variable (which is a factor, "active", "inactive"), because otherwise an error will be thrown out
|#str(distset) 
  distsetb <- as.data.frame(as.matrix(distset))
  str(distsetb)
  
  # Convert to a tibble
  library("tibble")
  distsetc <- as_data_frame(distsetb)
#  distsetc
#  str(distsetc)
  
  
  #:nrow(distsetc)
  distsetd <- t(apply(distsetc, 1, sort, decreasing = FALSE)) # each row is sorted in ascending order
# str(distsetd)
# distsetd[, 1] # it has only zero values
  distsete <- distsetd[, -1] # the first column, having only zero values, is eliminated
# distsete
  distsetf <- as.data.frame(distsete)
  
  
  # the subset of kvalue columns is selected, i.e. k=kvalue (the first kvalue neighbours for each substance)
  distsetg <- distsetf[, 1:kvalue]
# str(distsetg)
  distsetg$row_means <- rowMeans(distsetg)
  
  q3_chi <- quantile(distsetg$row_means, prob=0.75) # the third quartile
  q3_chi
  
  q1_chi <- quantile(distsetg$row_means, prob=0.25) # the first quartile
  q1_chi
  
  refkvalue <- q3_chi + 1.5*(q3_chi-q1_chi) # ref value for k=kvalue, in train set (according to the IQR rule, outlieris are beyond 1.5*IQR)
# refkvalue
  names(refkvalue) <- "refkvalue" 
# refkvalue
  
  distseth <- as.data.frame(distsetf <= refkvalue) # checking for all distance columns if they are < reference value
# str(distseth)
  distseth$suma <- rowSums(distseth) # computing the sum of distances lower than refkvalue
  distseth$suma
  
  distsetf$density <- distseth$suma
  distsetf$threshold <- 0
  
  
  library(Rfast)
  
  k <- which(colnames(distsetf)=="density")
  
  for (i in 1:length(distsetf$density)){
    if(distsetf[i, k]!=0){
      distsetf[i, "threshold"] <- sum(as.numeric(distsetf[i,])[1:distsetf[i,k]])/distsetf[i,k]}
    else
    {distsetf[i, "threshold"] <- 0}
  }
  
  # this loop sums all distances smaller than the ref. value and computes, 
# thus establishing the individual threshold value, t_index_i (ti) from the published paper

 # distsetf$threshold
  sort(distsetf$threshold)
  distsetf$threshold[which(distsetf$threshold==0)] <-  min(distsetf$threshold[distsetf$threshold!=0]) # assign for the zero value the minimum value different from 0 
  sort(distsetf$threshold)
  
  
  test_dist_euc <- dist(x=traindataset[, model_colnames], y=testdataset[, model_colnames], method=method, upper=TRUE) # euclidean (or other method) distances to the test data set
# test_dist_euc
# str(test_dist_euc)
  test_dist_euc2 <- as.data.frame.matrix(test_dist_euc) # ! to be able to convert this list of "crossdist" class
# str(test_dist_euc2)
  
  test_dist_euc2[1:length(test_dist_euc2)] # first column is not zero now
  test_dist_euc2$threshold <- distsetf$threshold # apply the same threshold
# str(test_dist_euc2)
  
  test_dist_euc3 <- test_dist_euc2[, 1:length(test_dist_euc2)]<=test_dist_euc2$threshold # check against the threshold
# test_dist_euc3
  colsums(test_dist_euc3)

  outliers <- which(colsums(test_dist_euc3)==0)
  print(outliers)
  return(outliers)
}


# ADomain_Roy: based on Roy, K.; Kar, S.; Ambure, P. On a simple approach for determining applicability domain of QSAR models. Chemom. Intell. Lab. Syst. 2015, 145, 22–29.

ADomain_Roy <- function(train_dataset, test_dataset, model_colnames){
  
  
  train_out2 <- train_dataset[, model_colnames] # model_colnames does not include the outcome (dependent variable) column
  str(train_out2)
  test_out2 <- test_dataset[, model_colnames]
  str(test_out2)
  
  ncoltrain <- length(train_out2) # number of columns of the training data set
  # ncoltrain
  
  ncoltest <- length(test_out2) # number of columns of the test data set
  # ncoltest
  
  means_train2 <- colMeans(train_out2) # computing the means of the columns (descriptors) from the training set
  # means_train2
  
  library(matrixStats)
  sds_train2 <- colSds(as.matrix(train_out2)) # computing the sd for the columns (descriptors) from the training set
  # sds_train2
  
  
  train_out2_std_a <-abs(train_out2 - means_train2[col(train_out2)]) # first step for standardization, subtracting the column mean for each column (and taking the absolute value)
  str(train_out2_std_a)  
  
  
  train_out2_std_b <- train_out2_std_a/sds_train2[col(train_out2_std_a)] # second step for standardization, dividing to the s.d. of each column
  # str(train_out2_std_b) # check the standardized values 
  # ((train_out2_std_a$ZM1Mad)/sd(train_out2$ZM1Mad))[1:10] # comparing these values with those from the line below,  to see if the same result is obtained
  # train_out2_std_b[, 1][1:10] # (they are the same, but checking with an equality test results in FALSE, probably because of rounding errors)
  
  
  # doing the same for the test_set
  
  test_out2_std_a <- abs(test_out2 - means_train2[col(test_out2)])
  str(test_out2_std_a)
  # abs(test_out2$ZM1Mad-mean(train_out2$ZM1Mad)) # manual checking by comparison with the next line
  # test_out2_std_a$ZM1Mad
  
  test_out2_std_b <- test_out2_std_a/sds_train2[col(test_out2_std_a)]
  str(test_out2_std_b)
  # (test_out2_std_a$ZM1Mad)/sd(train_out2$ZM1Mad) # manual checking by comparison with the next line
  # test_out2_std_b$ZM1Mad
  
  
  train_out2_std_b$Max <- rowMaxs(as.matrix(train_out2_std_b)) # computing Smax (the descriptor with the largest value) for each row
  # max(train_out2_std_b[1,]) # manual checking by comparing the first row with the first value returned by the next line
  # train_out2_std_b$Max[1]
  
  test_out2_std_b$Max <- rowMaxs(as.matrix(test_out2_std_b))
  max(test_out2_std_b[1,])
  test_out2_std_b$Max[1]
  
  train_out2_std_b$Min <- rowMins(as.matrix(train_out2_std_b[, 1:ncoltrain])) # computing Smin (the descriptor with the lowest value) for each row of the training set
  min(train_out2_std_b[1,]) # # manual checking by comparing the first row with the first value returned by the next line
  train_out2_std_b$Min[1]
  
  test_out2_std_b$Min <- rowMins(as.matrix(test_out2_std_b[, 1:ncoltest])) # computing Smin for each row of the test set
  min(test_out2_std_b[1,])
  test_out2_std_b$Min[1]
  
  train_out2_std_b$mean_sk <- rowMeans(as.matrix(train_out2_std_b[, 1:ncoltrain])) # computing mean_sk (equation 2 from Roy et al. publication) for the train set
  test_out2_std_b$mean_sk <- rowMeans(as.matrix(test_out2_std_b[, 1:ncoltrain])) # computing mean_sk (equation 2 from Roy et al. publication) for the train set
  
  train_out2_std_b$sigma_sk <- rowSds(as.matrix(train_out2_std_b[, 1:ncoltrain])) # computing mean_sk (equation 2)
  test_out2_std_b$sigma_sk <- rowSds(as.matrix(test_out2_std_b[, 1:ncoltrain]))  # computing mean_sk  (equation 2)
  
  train_out2_std_b$S_new <- train_out2_std_b$mean_sk+1.28*train_out2_std_b$sigma_sk # computing S_new (equation 2)
  train_out2_std_b$mean_sk[1] + 1.28*train_out2_std_b$sigma_sk[1]
  train_out2_std_b$S_new[1]
  
  test_out2_std_b$S_new <- test_out2_std_b$mean_sk+1.28*test_out2_std_b$sigma_sk # computing S_new (equation 2)
  test_out2_std_b$mean_sk[1] + 1.28*test_out2_std_b$sigma_sk[1]
  test_out2_std_b$S_new[1]
  
  
  train_out2_std_b$outlier <-0 # initializing an outlier column
  for (i in 1:nrow(train_out2_std_b)) {
    if (train_out2_std_b$Max[i] < 3) {train_out2_std_b$outlier[i] <- FALSE }
    else if (train_out2_std_b$Min[i] > 3) {train_out2_std_b$outlier[i] <- TRUE }
    else if (train_out2_std_b$S_new [i] >3) {train_out2_std_b$outlier[i] <- TRUE }
    else {train_out2_std_b$outlier[i] <- FALSE }
  } # applying the criteria required by the paper
  train_out2_std_b$outlier
  outlieri3_train <- which(train_out2_std_b$outlier==1)
  train_no_outlieri <- train_out2[-outlieri3_train, ]
  
  
  test_out2_std_b$AplicD <-0 # initializing a vector for values in or outside applicability domain
  for (i in 1:nrow(test_out2_std_b)) {
    if (test_out2_std_b$Max[i] < 3) {test_out2_std_b$AplicD[i] <- TRUE }
    else if (test_out2_std_b$Min[i] > 3) {test_out2_std_b$AplicD[i] <- FALSE }
    else if (test_out2_std_b$S_new [i] >3) {test_out2_std_b$AplicD[i] <- FALSE }
    else {test_out2_std_b$AplicD[i] <- TRUE }
  } # # applying the criteria required by the paper
  test_out2_std_b$AplicD
  
  outlieri3 <- which(test_out2_std_b$AplicD==0)
  outlieri3
  return(outlieri3)
  
}

# AD_KDEOS and AD_INFLO, based on the DDoutlier package and the relevant references cited there:

AD_KDEOS <- function(traindataset, testdataset, model_colnames, threshold){ # threshold: e.g. 0.05 for 5%
  outlier <- rep("", times=nrow(testdataset))
  i <- 1
  while (i <= nrow(testdataset)) 
  {
    print(i)
    library(dplyr)
    newdf <- bind_rows(traindataset[, model_colnames], testdataset[i, model_colnames])
    library(DDoutlier)
    outlier_score <- KDEOS(dataset=newdf, k_min=3, k_max=10)
    names(outlier_score) <- 1:nrow(newdf)
    outlier[i] <- sum(names(sort(outlier_score, decreasing = TRUE))[1:round(nrow(testdataset)*threshold)] %in% as.character(nrow(newdf)))
    # if the sum is 1, it means that i is an outlier (the score is among the threshold - e.g. 5% - of the scores)
    print(outlier[i])
    i <- i+1
  } 
  return(outlier)
}
# the row indices of the outliers are returned


AD_INFLO <- function(traindataset, testdataset, model_colnames, threshold){ # threshold: e.g. 0.05 for 5%
  outlier <- rep("", times=nrow(testdataset))
  i <- 1
  while (i <= nrow(testdataset)) 
  {
    print(i)
    library(dplyr)
    newdf <- bind_rows(traindataset[, model_colnames], testdataset[i, model_colnames])
    library(DDoutlier)
    outlier_score <- INFLO(dataset=newdf, k=5)
    names(outlier_score) <- 1:nrow(newdf)
    outlier[i] <- sum(names(sort(outlier_score, decreasing = TRUE))[1:round(nrow(testdataset)*threshold)] %in% as.character(nrow(newdf)))
    # daca suma este 1, inseamna ca exista un outlier (scorul se situeaza in primele 5% scoruri)
    print(outlier[i])
    i <- i+1
  } 
  return(outlier)
}

# the row indices of the outliers are returned





